/*
 * store_bf.h
 *
 * Internal declarations for the Butterfly cache_dir routines
 *
 * $Id: store_bf.h,v 1.1.2.7 2000/09/20 19:00:18 sizif Exp $
 */

#ifndef __STORE_BF_H__
#define __STORE_BF_H__

#include <linux/reiserfs_fs.h>
#include <linux/aio.h>


typedef struct reiserfs_raw_ioctl_arg reiserfs_raw_ioctl_arg;
typedef struct aiocb aiocb;

typedef struct _bf_stats bf_stats_t;
typedef struct _bfinfo_t bfinfo_t;
typedef struct _bfop_t bfop_t;
typedef struct _bfstate_t bfstate_t;


/* When read/write/close request comes for sio which hasn't fd
   available yet, we cannot fire up aio_* operation for it
   immediately.  Instead, we queue the requests and process them after
   open finishes.  The following structure is the element of the
   queues. */

enum bf_type {
	BFNONE, BFOPEN, BFCREAT, BFREAD, BFWRITE,
	BFCLOSE, BFUNLINK, BFSETATTR,
	BFMAX,
};
#define BFOP_NAMES {"none","open","create","read","write", \
                    "close","unlink","setattr"}

struct _bfop_t {
    bfop_t *next;		/* efficient dlist -- don't need xcalloc */
    bfop_t *prev;
    bfop_t *busychain;
    enum bf_type type;
    storeIOState *sio;		/* our owner */
    aiocb aio;
    union {
	STRCB *read;
	FREE *write;		/* free_func */
    } callback;
    void *callback_data;	/* only for callback.read */
    off_t rlen;			/* real real/write length, syscall return */
    int starttime;		/* start timestamp, ms */
    unsigned int started:1;	/* 0 until launched */
    unsigned int finished:1;	/* 0 until landed */
};

struct _bfinfo_t {
    int dir;			/* fd of sd->path for reiserfs_raw ioctls */
    link_list *kicklater;	/* list of bfstates that got EAGAIN */
    int refuse_swapout;		/* storeBfCreate returns NULL if more
				   than this ops away */
};

struct _bfstate_t {
    int fd;			/* file fd, -1 if not opened yet */
    reiserfs_raw_ioctl_arg ra;	/* single ra shared by all parallel ops */
    bfop_t ops;			/* op used for BF{OPEN,CREAT,UNLINK,SETATTR}
				   and as queue head for BF{READ,WRITE} */
    int errflag;		/* error that caused abortion of the sio.
				   Normaly zero and no abort.  */
    unsigned int close_request:1; /* close when q is empty */
    unsigned int close_started:1; /* BF_CLOSE launched already */
    unsigned int aborting:1;	  /* close ASAP (flush not run the q) */
    unsigned int unlink_attempted:1; /* already tried to unlink the file */
};

struct _bf_stats {
    unsigned eagain;
    unsigned away;
    unsigned max_away;
    unsigned abs_max_away;
    unsigned swapout_refused;
    struct bf_stats_counters {
	unsigned ops;
	unsigned success;
	unsigned fail;
	unsigned max_dura;	/* maximal op duration */
	float    avg_dura;	/* average op duration */
	float    avg_dura100;	/* averege op duration in last 100 ops */
    } counters[BFMAX];
};

/* FIFO lists of launched ops.  We keep different ops in different
   lists because op delays and urgency are different */
struct bf_busyops {
    bfop_t *head;		/* FIFO list of launched BF of certain type */
    bfop_t **tail;	 	/* tail of the FIFO list */
};

extern MemPool *bf_state_pool;
extern MemPool *bf_op_pool;
extern bf_stats_t bf_stats;
extern struct bf_busyops bf_busyops[BFMAX];
extern char *op_names;
extern FILE *bf_opstats;		/* bulky binary op stats file */
extern char *bf_opstats_fn;


/* from store_io_bf.c */
extern int   storeBfKickQueue (bfstate_t *bfstate);
extern void  storeBfOpCompletion (bfop_t *op);

/* from store_dir_bf.c */
extern void  storeBfDirUnrefObj(SwapDir * SD, StoreEntry * e);
extern void  storeBfDirReplAdd(SwapDir * SD, StoreEntry * e);
extern void  storeBfDirReplRemove(StoreEntry * e);
extern void  storeBfDirFillRa(StoreEntry *e, reiserfs_raw_ioctl_arg *ra);

/*
 * Store IO stuff
 */
extern STOBJCREATE storeBfCreate;
extern STOBJOPEN storeBfOpen;
extern STOBJCLOSE storeBfClose;
extern STOBJREAD storeBfRead;
extern STOBJWRITE storeBfWrite;
extern STOBJUNLINK storeBfUnlink;


/* Get microsecond timestamp for profiling.  Almost verbatim from
   linux/mm/filemap.c (KAIO) */

static inline unsigned long
kaio_time(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000 + tv.tv_usec);
}


#endif /* __STORE_BF_H__ */
