
/*
 * $Id: store_dir_bf.c,v 1.1.2.14 2000/09/20 19:00:18 sizif Exp $
 *
 * DEBUG: section 86    Store Directory, the "Butterfly" version
 * AUTHOR: Yury Shevchuk <sizif@botik.ru>
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by the
 *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
 *  the Regents of the University of California.  Please see the
 *  COPYRIGHT file for full details.  Squid incorporates software
 *  developed and/or copyrighted by other sources.  Please see the
 *  CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 */

#include "squid.h"
#include <signal.h>

#if HAVE_STATVFS
#if HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif
#endif

#include "store_bf.h"

bf_stats_t bf_stats;

typedef struct _RebuildState RebuildState;
struct _RebuildState {
    SwapDir *sd;
    struct _store_rebuild_data counts;
    int clean;
};

int n_bf_dirs = 0;
static int *bf_dir_index = NULL;
MemPool *bf_state_pool = NULL;
MemPool *bf_op_pool = NULL;
static int bf_initialised = 0;
static int bf_nthreads = 0;
struct bf_busyops bf_busyops[BFMAX];
char *bfop_names[] = BFOP_NAMES;
FILE *bf_opstats;		/* bulky binary op stats file */
char *bf_opstats_fn;

static int  storeBfDirCreateDirectory(const char *path,
					     int should_exist);
static int  storeBfDirVerifyDirectory(const char *path);
static int  storeBfDirVerifyCacheDirs(SwapDir * sd);
static void storeBfDirOpenSwapLog(SwapDir * sd);
static void storeBfDirCloseSwapLog(SwapDir * sd);
static void storeBfDirInit(SwapDir * sd);
static void storeBfStats(StoreEntry * sentry);
static void storeBfDirSync(SwapDir * SD);
static int  storeBfDirCallback(SwapDir * SD);
static void storeBfRelease(StoreEntry *e);
static SwapDir *storeBfReselectDir (const StoreEntry *e);
static void storeBfDirRebuild(SwapDir * sd);
static int storeBfDirWriteCleanStart(SwapDir * sd);
static const StoreEntry *storeBfDirCleanLogNextEntry(SwapDir * sd);
static void storeBfDirWriteCleanDone(SwapDir * sd);
static void storeBfDirSwapLog(const SwapDir * sd, const StoreEntry * e, int op);
static void storeBfDirNewfs(SwapDir * sd);
static int  storeBfDirIs(SwapDir * sd);
static void storeBfDirMaintain(SwapDir * SD);
static int  storeBfSelectSwapDir(const StoreEntry *);
static void storeBfDirRefObj(SwapDir * SD, StoreEntry * e);
static void storeBfDirStats(SwapDir * SD, StoreEntry * sentry);
static void storeBfDirReconfigure(SwapDir * sd, int index, char *path);
static void storeBfDirDump(StoreEntry * entry, const char *name, SwapDir * s);
static void storeBfDirFree(SwapDir * s);
static int  storeBfCleanupDoubleCheck(SwapDir * sd, StoreEntry * e);
static void storeBfDirParse(SwapDir * sd, int index, char *path);


int
sanitycheck()
{
    int t;
    unsigned true_away = 0;

    for (t = BFNONE+1; t < BFMAX; t++) {
	bfop_t *op = bf_busyops[t].head;
	bfop_t **true_tail = &bf_busyops[t].head;
	while (op) {
	    true_away++;
	    true_tail = &op->busychain;
	    op = op->busychain;
	}
	assert (bf_busyops[t].tail == true_tail);
    }

    assert (bf_stats.away == true_away);
    return 0;
}

static int
storeBfDirCreateDirectory(const char *path, int should_exist)
{
    /* our directories are root dirs of reiserfs_raw fs's.  Never
       create anything. */
    if (storeBfDirVerifyDirectory (path) != 0)
	fatalf("Swap directory %s needs manual intervention\n", path);
    return 0;
}

static int
storeBfDirVerifyDirectory(const char *path)
{
    struct stat sb;
    reiserfs_raw_ioctl_arg ra;
    int fd, fd2;

    if (stat(path, &sb) < 0) {
	debug(20, 0) ("%s: %s\n", path, xstrerror());
	return -1;
    }
    if (S_ISDIR(sb.st_mode) == 0) {
	debug(20, 0) ("%s is not a directory\n", path);
	return -1;
    }
    fd = open (path, O_RDONLY);
    if (fd < 0) {
	debug(20, 0) ("%s is unreadable for uid %u, gid %u\n", path);
	return -1;
    }
    ra.id1 = ra.id2 = 7;
    fd2 = ioctl (fd, REISERFS_IOC_RAWOPEN, &ra);
    if (fd2 == -1 && errno != ENOENT) {
	debug(20, 0) ("%s is not a REISERFS-RAW directory\n", path);
	close(fd);
	return -1;
    }
    close(fd2);
    close(fd);
    return 0;
}

/*
 * This function is called by storeBfDirInit().  If this returns < 0,
 * then Squid exits, complains about swap directories not
 * existing, and instructs the admin to run 'squid -z'
 */
static int
storeBfDirVerifyCacheDirs(SwapDir * sd)
{
    const char *path = sd->path;
    if (storeBfDirVerifyDirectory(path) < 0)
	return -1;
    return 0;
}

static void
storeBfDirOpenSwapLog(SwapDir * sd)
{
    if (0 == n_bf_dirs)
	assert(NULL == bf_dir_index);
    n_bf_dirs++;
    assert(n_bf_dirs <= Config.cacheSwap.n_configured);
}

static void
storeBfDirCloseSwapLog(SwapDir * sd)
{
    n_bf_dirs--;
    n_storetree_dirs--;
    assert(n_bf_dirs >= 0);
    if (0 == n_bf_dirs)
	safe_free(bf_dir_index);
}

static void
storeBfDirInit(SwapDir * sd)
{
    bfinfo_t *bfi = sd->fsdata;
    static const char *errmsg =
    "\tFailed to verify one of the swap directories, Check cache.log\n"
    "\tfor details.  Run 'squid -z' to create swap directories\n"
    "\tif needed, or if running Squid for the first time.";

    if (!getenv(AIO_THREAD_ENVIRON)) {
	static char aio_nthreads[30];
	snprintf(aio_nthreads, sizeof(aio_nthreads), "%d", bf_nthreads);
	setenv(AIO_THREAD_ENVIRON, aio_nthreads, 0);
    }
    if (storeBfDirVerifyCacheDirs(sd) < 0)
	fatal(errmsg);

    bfi->dir = open (sd->path, O_RDONLY);
    if (bfi->dir < 0) {
	debug(50, 0) ("storeBfInit: open(%s): %s\n", sd->path, xstrerror());
	fatal ("cache_dir open failed");
    }
#if 0				/* TODO: consider legalizing dir fd? */
    fd_note(bfi->dir, sd->path);
    commSetTimeout(bfi->dir, -1, NULL, NULL);
    commSetNonBlocking(bfi->dir);
#endif
    /* We are doing polled I/O, don't wait too long in comm_poll */
    comm_quick_poll_required();
    storeBfDirOpenSwapLog(sd);
    storeBfDirRebuild(sd);
}


static void
storeBfStats(StoreEntry * sentry)
{
    int t;

    if (bf_stats.abs_max_away < bf_stats.max_away)
	bf_stats.abs_max_away = bf_stats.max_away;
    storeAppendPrintf(sentry, "AIO ops in flight: now %u max %u absmax %u\n",
		      bf_stats.away,
		      bf_stats.max_away,
		      bf_stats.abs_max_away);
    bf_stats.max_away = 0;

    storeAppendPrintf(sentry, "EAGAIN errors: %u\n", bf_stats.eagain);
    storeAppendPrintf(sentry, "Swapout refused because of disk overload: %u\n",
		      bf_stats.swapout_refused);
    storeAppendPrintf(sentry, "\n            OPS SUCCESS    FAIL"
		      "   DURATION, ms: MAX   AVG  AVG-for-last-100-ops\n");
                                       
    for (t = BFNONE+1; t < BFMAX; t++) {
	struct bf_stats_counters *c = &bf_stats.counters[t];
	storeAppendPrintf(sentry, "%7s %7u %7u %7u                %5u %5.0f %5.0f\n",
			  bfop_names[t],
			  c->ops, c->success, c->fail,
			  c->max_dura, c->avg_dura, c->avg_dura100);
    }

    sanitycheck();
}

/*
 * storeBfDirSync
 *
 * Sync any pending data. We just sit around and read the queue
 * until the data has finished writing.
 */
static void
storeBfDirSync(SwapDir * sd)
{
  /* later. */
#if 0
    bfinfo_t *bfi = sd->fsdata;
    while (bfi->away > 0) {
	debug(86, 1) ("storeBfDirSync: %d messages away\n", bfi->away);
	storeBfDirCallback(SD);
    }
#endif
}

/* Write down op duration profiling data */

static void
storeBfWriteOpStats (bfop_t *op)
{
    if (bf_opstats_fn && ! bf_opstats) {
	bf_opstats = fopen (bf_opstats_fn, "a");
	if (! bf_opstats) {
	    debug(50, 1) ("WARNING: %s: %s\n", bf_opstats_fn, xstrerror());
	    bf_opstats_fn = 0;
	}
    }
    fwrite (&op->type, sizeof (op->type), 1, bf_opstats);
    fwrite (&op->aio.aio_times, sizeof (op->aio.aio_times), 1, bf_opstats);
}


/*
 * storeBfDirCallback
 *
 * This is called regularly from the main select() loop to let us serve
 * aio completion requests.
 */
static int
storeBfDirCallback(SwapDir * sd)
{
    bfinfo_t *bfi = sd->fsdata;
    storeIOState *sio;
    static struct timeval last_call;
    int count_completions = 0;
    int count_kicks = 0;
    int i;

    /* Queue lookup order.  First, serve ops that require no further
       processing (their processing doesn't result in growth of the
       queue).  BFOPEN and BFCREAT trigger bursts of new messages, so
       they come last.  BFOPEN before BFCREAT as open is "interactive"
       while creat is "batch" (swap-out delay does not matter) */
    static int op_order[] = {BFCLOSE, BFUNLINK, BFREAD, BFWRITE,
			     BFOPEN, BFCREAT, BFNONE};
    int scan_depth[] = {3, 3, 30, 3, bf_nthreads, 1};

    /* No point in scanning queues more often than once in 3ms */
    if ((unsigned long)tvSubUsec(last_call, current_time) < 3000)
	return 0;

    last_call = current_time;
    
    if (bf_stats.away > 0)
	debug(86,3) ("storeBfDirCallback: away %u\n", bf_stats.away);

    for (i = 0; op_order[i] != BFNONE; i++) {
	bfop_t *op;
	int t = op_order[i];
	int depth = scan_depth[i]; /* go through this many non-ready ops */
	bfop_t **opp = &bf_busyops[t].head;

	while ((op = *opp)) {
	    if (aio_error(&op->aio) == EINPROGRESS) {
		debug (86,3) ("storeBfDirCallback: %p incomplete, depth %d\n", op, depth);
		opp = &op->busychain;
		if (--depth <= 0)
		    break;
	    }
	    else {
		*opp = op->busychain;
		if (! *opp)
		    bf_busyops[t].tail = opp;
		bf_stats.away--;
		debug (86,3) ("storeBfDirCallback: %p complete\n", op);
		if (bf_opstats_fn) {
		    op->aio.aio_times[AIOTIME_FINISH] = kaio_time ();
		    storeBfWriteOpStats (op);
		}
		storeBfOpCompletion (op);
		count_completions++;
	    }
	}
    }

    while ((sio = linklistShift (&bfi->kicklater))) {
	if (cbdataValid (sio)) {
	    bfstate_t *bfstate = sio->fsstate;
	    if (storeBfKickQueue (bfstate) < 0) {
		/* EAGAIN again.  Further kick attempts will likely
		   fail too, so stop for now (busy wait processing the
		   same sio isn't the best we can do). */
		cbdataUnlock (sio);
		break;
	    }
	    count_kicks++;	/* unsucessful uncounted */
	}
	cbdataUnlock (sio);
    }

    if (count_completions + count_kicks > 0)
	debug (86,3) ("storeBfDirCallback: done %u completions, %u kicks\n",
		      count_completions, count_kicks);

    return count_completions + count_kicks;
}


/* Remove StoreEntry from memory while keeping the disk object.  Save
   important e->lastref and e->expires info in the file's stat data,
   so expiration process (storeTreeDirClean) can make decisions based
   on them without reading the file.

   For reiserfs_raw, we avoid setting attributes now.  They has been
   set on object creation, and the only change we may need now is
   update lastref attribute which is currently set to the time of file
   creation or last open.  (Hm, we must update lastref for hot objects
   periodically somewhere, or they might get GCed on the disk.)  */

static void
storeBfRelease(StoreEntry *e)
{
    assert (e->lock_count == 0);

    if (EBIT_TEST (e->flags, ENTRY_TENTATIVE)) {
	/* If we are asked to release tentative entry, this means it
	   wasn't succesful (no file found for this entry).  It is no
	   longer tentative, it is simply dummy StoreEntry. */
	debug(86, 3) ("storeBfRelease: rm unsuccesful tentative entry %p\n", e);
	goto kill_it;
    }

#define M(BITNUM) (1L<<(BITNUM))
    if ((e->flags &~ M(ENTRY_VALIDATED)) != (M(ENTRY_CACHABLE)|M(ENTRY_DISPATCHED))) {
	/* Uncommon, leave in memory */
	debug(86, 3) ("storeBfRelease: leaving in memory entry with uncommon flags %s\n",
		      storeEntryFlags(e));
	return;
    }

 kill_it:
    storeSetMemStatus(e, NOT_IN_MEMORY);
    destroy_StoreEntry(e);
}


/* This is called by store dir selection procedure to let us redirect
   the object to one of our fellow bf dirs --
   storeBfSelectSwapDir will say which one using a hash based on
   URL */

static SwapDir *
storeBfReselectDir (const StoreEntry *e)
{
    return INDEXSD(storeBfSelectSwapDir(e));
}


static void
storeRebuildDummy(void *data)
{
    RebuildState *rb = data;

    /*
     * Everything already done in storeBfDirRebuild.  But calling
     * storeRebuildComplete from there is banned.
     */
    store_dirs_rebuilding--;
    storeRebuildComplete(&rb->counts);
    cbdataFree(rb);
}

CBDATA_TYPE(RebuildState);

static void
storeBfDirRebuild(SwapDir * sd)
{
    RebuildState *rb = xcalloc(1, sizeof(*rb));
    EVH *func = storeRebuildDummy;

    CBDATA_INIT_TYPE(RebuildState);
    rb = CBDATA_ALLOC(RebuildState, NULL);
    rb->sd = sd;
    store_dirs_rebuilding++;
    eventAdd("storeRebuild", func, rb, 0.0, 0);
}


/*
 * Begin the process to write clean cache state.  For BF this
 * is no-op.
 */
static int
storeBfDirWriteCleanStart(SwapDir * sd)
{
    sd->log.clean.write = NULL;
    sd->log.clean.state = NULL;
    return 0;  /* -1 to indicate error */
}

/*
 * Get the next entry that is a candidate for clean log writing
 */
static const StoreEntry *
storeBfDirCleanLogNextEntry(SwapDir * sd)
{
    return NULL;
}

static void
storeBfDirWriteCleanDone(SwapDir * sd)
{
    /* sync (wait while diskd finishes) and write down age classes.  later */
}

static void
storeBfDirSwapLog(const SwapDir * sd, const StoreEntry * e, int op)
{
}    

static void
storeBfDirNewfs(SwapDir * sd)
{
    debug(47, 3) ("Creating swap space in %s\n", sd->path);
    storeBfDirCreateDirectory(sd->path, 0);
    /* storeBfDirCreateSwapSubDirs(sd); */
}


/* We use 64bit version of r5 hash to make reiserfs_raw file id.  R5
   it chosen for its property of more or less preserving original data
   order, which will hopefully be good for seek times.

   The code has been copied from linux/fs/reiserfs/hashes.c, the only
   change is s/u32/unsigned long long/g */

static unsigned long long
r5l_hash (const char *msg, int len)
{
    unsigned long long a = 0;
    while (*msg) { 
	a += *msg << 4;
	a += *msg >> 4;
	a *= 11;
	msg++;
    } 
    return a;
}

static inline unsigned long long
make_id (const char *key, const char *msg, int len)
{
#if 1
    return r5l_hash (msg, len);
#else
    return *(unsigned long long *)(&key[8]);
#endif
}



/* This is important function called from storeGetPublic,
   disk-oriented equivalent of storeGet.  */

StoreEntry *
storeTreeGet (const char *url, const cache_key *key)
{
    StoreEntry *e;

    if (n_storetree_dirs == 0)
	return NULL;
    if (strncasecmp (url, "cache_object://", sizeof "cache_object://"-1) == 0)
	return NULL;

    debug(86, 5) ("StoreTreeGet for %s %s  Callers %p %p %p %p\n", url,
		  storeKeyText (key),
		  __builtin_return_address(0),
		  __builtin_return_address(1),
		  __builtin_return_address(2),
		  __builtin_return_address(3)
		  );

    /* We don't know if the requested object is in the store, and we
       don't want to waste time on stat(2) to find out.  So we create
       tentative StoreEntry with requested URL as if it were in the
       store, and let upper layers try to open it.  They will try to
       swap in and if the object is not on the disk, they's get open
       error and handle the error -- for example, in clientCacheHit
       they jump to clientCacheMiss on swapin error. */

    e = new_StoreEntry (STORE_ENTRY_WITHOUT_MEMOBJ, NULL, NULL);
    e->store_status = STORE_OK;
    e->mem_status = NOT_IN_MEMORY;
    e->swap_status = SWAPOUT_DONE;
    e->swap_filen = 73;		/* for luck */
    /* would have to read in the object to restore the flags.  To
       avoid that, we'll teach storeTreeRelease not to purge
       StoreEntries with non-common flags (ENTRY_NEGCACHED, for one) */
    e->flags = 0;
    EBIT_SET(e->flags, ENTRY_CACHABLE);
    EBIT_SET(e->flags, ENTRY_DISPATCHED);
    EBIT_SET(e->flags, ENTRY_VALIDATED);
    EBIT_SET(e->flags, ENTRY_TENTATIVE);
    e->ping_status = PING_NONE;
    storeHashInsert (e, key);

    e->id = make_id (key, url, strlen (url));
    e->id_valid = 1;
    e->swap_dirn = storeBfSelectSwapDir (e);

    return e;
}


/* Restore StoreEntry fields from metadata just read from disk object.
   Called from storeClientReadHeader */
void
storeTreeRestoreMetadata(StoreEntry *e, tlv *t)
{
    SwapDir *sd = INDEXSD(e->swap_dirn);
    size_t save_sz = e->swap_file_sz;

    assert(t->type == STORE_META_STD);
    assert(t->length == STORE_HDR_METASIZE);
    assert(storeBfDirIs (sd));
    xmemcpy(&e->timestamp, t->value, STORE_HDR_METASIZE);

    /* swap_file_sz in the metadata is most likely 0, it is not known
       at the point when swap-out starts.  Replace with known good
       value that we got from RAWOPEN */
    e->swap_file_sz = save_sz;

    /* this will probably be set later, but just in case */
    e->lastref = squid_curtime;
}


static int
storeBfDirIs(SwapDir * sd)
{
    return (strcmp(sd->type, "butterfly") == 0);
}

static void
storeBfDirMaintain(SwapDir * SD)
{
    /* reiresfs-raw has object removal implemented in the kernel. */
    return;
}

/*
 * storeBfDirCheckObj
 *
 * This routine is called by storeDirSelectSwapDir to see if the given
 * object is able to be stored on this filesystem.  (Round-robin dir
 * selection alg doesn't use this at all).
 *
 */
int
storeBfDirCheckObj(SwapDir * SD, const StoreEntry * e)
{
    if (storeBfSelectSwapDir(e) != SD->index)
        return -1;
    return 0;
}


/* Select the cachedir where to put/where to look
   for the given entry.  The current implementation is
   quick&dirty: assumes all bf dirs are of equal size.
   The right way is to weight based on dir sizes, TODO later.
   Look at CARP as Henrik suggested?  */

static int
storeBfSelectSwapDir(const StoreEntry * e)
{
    int dirn;
    int x;

    if (! e->id_valid) {
	assert (e->mem_obj);
	assert (e->mem_obj->url);
	e->id = make_id (e->key, e->mem_obj->url, strlen (e->mem_obj->url));
	e->id_valid = 1;
    }

    x = *(unsigned *)e->key % n_bf_dirs;
    
    if (Config.cacheSwap.n_configured == n_bf_dirs)
	return x;
    else for (dirn = 0; dirn < Config.cacheSwap.n_configured; dirn++) {
	SwapDir *sd = INDEXSD(dirn);
	if (storeBfDirIs(sd)) {
	    if (x == 0)
		return dirn;
	    x--;
	}
    }
    assert (0);
    return 0;
}


/*
 * storeBfDirRefObj
 *
 * This routine is called whenever an object is referenced, so we can
 * maintain replacement information within the storage fs.
 *
 * TODO put reiserfs_raw setattr here!
 */
static void
storeBfDirRefObj(SwapDir *sd, StoreEntry * e)
{
    debug(1, 3) ("storeBfDirRefObj: referencing %p %d/%d\n", e);
#if 0
    /* TODO: call storeBfSettattr here */
    if (sd->repl->Referenced)
	sd->repl->Referenced(sd->repl, e, &e->repl);
#endif
}

/*
 * storeBfDirUnrefObj
 * This routine is called whenever the last reference to an object is
 * removed, to maintain replacement information within the storage fs.
 */
void
storeBfDirUnrefObj(SwapDir * SD, StoreEntry * e)
{
    debug(1, 3) ("storeBfDirUnrefObj: unreferencing %p\n", e);
#if 0
    if (sd->repl->Dereferenced)
	sd->repl->Dereferenced(sd->repl, e, &e->repl);
#endif
}


/*
 * Add and remove the given StoreEntry from the replacement policy in
 * use.
 */

void
storeBfDirReplAdd(SwapDir * sd, StoreEntry * e)
{
#if 0
    debug(20, 4) ("storeBfDirReplAdd: added node %p to dir %d\n", e, sd->index);
    sd->repl->Add(sd->repl, e, &e->repl);
#endif
}


void
storeBfDirReplRemove(StoreEntry * e)
{
#if 0
    SwapDir *SD;
    if (e->swap_dirn < 0)
	return;
    SD = INDEXSD(e->swap_dirn);
    debug(20, 4) ("storeBfDirReplRemove: remove node %p from dir %d\n", e,
	SD->index);
    SD->repl->Remove(SD->repl, e, &e->repl);
#endif
}



static void
storeBfDirStats(SwapDir *sd, StoreEntry * sentry)
{
    bfinfo_t *bfi;
#if HAVE_STATVFS
    struct statvfs sfs;
#endif
    bfi = sd->fsdata;
#if HAVE_STATVFS
#define fsbtoblk(num, fsbs, bs) \
    (((fsbs) != 0 && (fsbs) < (bs)) ? \
            (num) / ((bs) / (fsbs)) : (num) * ((fsbs) / (bs)))
    if (!statvfs(sd->path, &sfs)) {
	storeAppendPrintf(sentry, "Filesystem Space in use: %d/%d KB (%d%%)\n",
	    fsbtoblk((sfs.f_blocks - sfs.f_bfree), sfs.f_frsize, 1024),
	    fsbtoblk(sfs.f_blocks, sfs.f_frsize, 1024),
	    percent(sfs.f_blocks - sfs.f_bfree, sfs.f_blocks));
    }
#endif

    storeAppendPrintf(sentry, "Flags:");
    if (sd->flags.selected)
	storeAppendPrintf(sentry, " SELECTED");
    if (sd->flags.read_only)
	storeAppendPrintf(sentry, " READ-ONLY");
    storeAppendPrintf(sentry, "\n");
}

/*
 * storeBfDirReconfigure
 *
 * This routine is called when the given swapdir needs reconfiguring 
 */
static void
storeBfDirReconfigure(SwapDir * sd, int index, char *path)
{
    char *token;
    unsigned int read_only = 0;
    char *opstats_fn = NULL;
    bfinfo_t *bfi = sd->fsdata;

    /* TODO -- remove RO or make it work.  See selectSwapDir. */

    while ((token = strtok(NULL, w_space))) {
	/* TODO -- remove RO or make it work.  See selectSwapDir. */
	if (!strcasecmp(token, "read-only"))
	    read_only = 1;
	else if (!strncasecmp(token, "opstats=", 8))
	    opstats_fn = xstrdup (token+8);
	else if (!strncasecmp(token, "refuse_swapout=", 15))
	    bfi->refuse_swapout = atoi(token+15);
    }

    /* just reconfigure it */
    if (sd->flags.read_only != read_only)
	debug(3, 1) ("Cache dir '%s' now %s\n",
	    path, read_only ? "Read-Only" : "Read-Write");
    sd->flags.read_only = read_only;

    if (bf_opstats) {
	fclose (bf_opstats);
	bf_opstats = NULL;
    }
    if (bf_opstats_fn)
	xfree (bf_opstats_fn);
    bf_opstats_fn = opstats_fn;

    return;
}

static void
storeBfDirDump(StoreEntry * entry, const char *name, SwapDir * sd)
{
    storeAppendPrintf(entry, "%s %s %s\n",
		      name,
		      "butterfly",
		      sd->path);
}

/*
 * Only "free" the filesystem specific stuff here
 */
static void
storeBfDirFree(SwapDir * sd)
{
    bfinfo_t *bfi = sd->fsdata;
    xfree(bfi);
    sd->fsdata = NULL;		/* Will aid debugging... */
}

void
storeBfDirFillRa(StoreEntry *e, reiserfs_raw_ioctl_arg *ra)
{
    if (! e->id_valid) {
	assert (e->mem_obj);
	assert (e->mem_obj->url);
	e->id = make_id (e->key, e->mem_obj->url, strlen (e->mem_obj->url));
	e->id_valid = 1;
    }
    ra->id1 = (unsigned) e->id;
    ra->id2 = (unsigned) (e->id >> 32);
    ra->lastref = e->lastref;
    ra->expires = ~0;		/* end of Epoch -- don't let kernel
                                   remove it for now. */
    ra->user1 = 0;		/* put headersz here, later */
}

/*
 * storeBfCleanupDoubleCheck
 *
 * This is called by storeCleanup() if -S was given on the command line.
 */
static int
storeBfCleanupDoubleCheck(SwapDir * sd, StoreEntry * e)
{
    /* very good dir, trust me! */
    return 0;
}

/*
 * storeBfDirParse
 *
 * Called when a *new* fs is being setup.
 */
static void
storeBfDirParse(SwapDir * sd, int index, char *path)
{
    char *token;
    unsigned int read_only = 0;
    unsigned int nthreads = 8;
    bfinfo_t *bfi;

    sd->fsdata = bfi = xcalloc(1, sizeof(*bfi));
    bfi->refuse_swapout = 1500;	/* default */

    while ((token = strtok(NULL, w_space))) {
	/* TODO -- remove RO or make it work.  See selectSwapDir. */
	if (!strcasecmp(token, "read-only"))
	    read_only = 1;
	else if (!strncasecmp(token, "threads=", 8))
	    nthreads = atoi(token+8);
	else if (!strncasecmp(token, "opstats=", 8)) {
	    if (bf_opstats_fn)
		debug(86, 1) ("WARNING: %s: already have an opstats file (%s), ignoring\n",
			      token, bf_opstats_fn);
	    else
		bf_opstats_fn = xstrdup (token+8);
	}
	else if (!strncasecmp(token, "refuse_swapout=", 15))
	    bfi->refuse_swapout = atoi(token+15);
    }

    sd->index = index;
    sd->path = xstrdup(path);
    /* keep cache_swap >= cache_mem check happy.  Use a constant for
       now, do it with statvfs later. */
    sd->max_size = 1000<<20;	/* 1 TB */
    sd->low_size = sd->max_size;
    sd->high_size = sd->max_size;
    sd->flags.read_only = read_only;
    sd->init = storeBfDirInit;
    sd->newfs = storeBfDirNewfs;
    sd->dump = storeBfDirDump;
    sd->freefs = storeBfDirFree;
    sd->dblcheck = storeBfCleanupDoubleCheck;
    sd->statfs = storeBfDirStats;
    sd->maintainfs = storeBfDirMaintain;
    sd->checkobj = storeBfDirCheckObj;
    sd->refobj = storeBfDirRefObj;
    sd->unrefobj = storeBfDirUnrefObj;
    sd->callback = storeBfDirCallback;
    sd->sync = storeBfDirSync;
    sd->halfrelease = storeBfRelease;
    sd->reselectdir = storeBfReselectDir;
    sd->obj.create = storeBfCreate;
    sd->obj.open = storeBfOpen;
    sd->obj.close = storeBfClose;
    sd->obj.read = storeBfRead;
    sd->obj.write = storeBfWrite;
    sd->obj.unlink = storeBfUnlink;
    sd->log.open = storeBfDirOpenSwapLog;
    sd->log.close = storeBfDirCloseSwapLog;
    sd->log.write = storeBfDirSwapLog;
    sd->log.clean.start = storeBfDirWriteCleanStart;
    sd->log.clean.nextentry = storeBfDirCleanLogNextEntry;
    sd->log.clean.done = storeBfDirWriteCleanDone;
    bf_nthreads += nthreads;

    n_storetree_dirs++;
}

/*
 * Initial setup / end destruction
 */
void
storeBfDirDone(void)
{
    memPoolDestroy(bf_state_pool);
    memPoolDestroy(bf_op_pool);
    if (bf_opstats)
	fclose (bf_opstats);
    if (bf_opstats_fn)
	xfree(bf_opstats_fn);
    bf_initialised = 0;
}

/* Setup routine.  This must be named by its long official name, not
   just _bf */

void
storeFsSetup_butterfly (storefs_entry_t * storefs)
{
    int t;

    assert(!bf_initialised);
    storefs->parsefunc = storeBfDirParse;
    storefs->reconfigurefunc = storeBfDirReconfigure;
    storefs->donefunc = storeBfDirDone;
    bf_state_pool = memPoolCreate("BF IO State data", sizeof(bfstate_t));
    bf_op_pool = memPoolCreate("BF IO operation data", sizeof(bfop_t));
    memset(&bf_stats, '\0', sizeof(bf_stats));
    cachemgrRegister("butterfly", "Butterfly Stats", storeBfStats, 0, 1);
    debug(81, 1) ("Butterfly started\n");
    for (t = 0; t < BFMAX; t++) {
	bf_busyops[t].head = NULL;
	bf_busyops[t].tail = &bf_busyops[t].head;
    }
    bf_initialised = 1;
}
