
/* $Id: store_io_bf.c,v 1.1.2.12 2000/09/20 19:00:18 sizif Exp $
 *
 * DEBUG: section 81   Butterfly cache_dir I/O functions.
 * Written by Yury Shevchuk <sizif@botik.ru>
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by the
 *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
 *  the Regents of the University of California.  Please see the
 *  COPYRIGHT file for full details.  Squid incorporates software
 *  developed and/or copyrighted by other sources.  Please see the
 *  CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 */

/*
 * Operation outline
 * ~~~~~~~~~~~~~~~~~
 * 
 * There are two groups of operations.
 * 
 * 1) Stateless:
 * 
 *    storeBfUnlink/Setattr
 * 
 * These are simple one-shot operations.  They don't have sio associated
 * with them.  storeBfUnlink/Setattr launches the required aio operation,
 * and storeBfOpCompletion cleanups allocated control data upon completion.
 * 
 * 2) Stateful.  The sequence goes like this:
 * 
 *    storeBfOpen/Creat
 *    storeBfRead/Write, storeBfRead/Write, ...
 *    storeBfClose
 * 
 * storeBfOpen/Create create sio and associated bfstate, launch
 * open/creat aio, and return (although the open is not completed yet).
 * The upper layers can immediately call storeBfRead/Write which we
 * cannot lauch right away as we don't have fd until Open/Creat
 * completes.  So we queue the request; there's a single queue for that
 * in struct bfstate.  The upper layers continue to supply requests using
 * storeBfRead/Write.
 * 
 * The queue is ran by storeBfKickQueue which is called from
 * storeBfOpCompletion (and also from storeBfRead/Write upon queueing
 * another request).
 * 
 * When storeBfClose is called, issuing the actual close request is
 * deferred until the queue is empty.
 * 
 */

#include "config.h"
#include "squid.h"

#include "store_bf.h"


static int   storeBfUnlink2(SwapDir *sd, StoreEntry *e, storeIOState *sio);
static int   storeBfLaunch (SwapDir *sd, int (*aiofunc)(aiocb *), bfop_t *op);
static void  storeBfTryLaunchClose (storeIOState *sio);
static void  storeBfCloseLastResort (storeIOState *sio);
static void  storeBfIOCallback(storeIOState * sio, int errflag);
static void  storeBfIOAbort (storeIOState *sio, int op_type, int errflag);
static void  storeBfOpEnqueue (bfstate_t *bfstate, bfop_t *op);
static void  storeBfSweep (bfstate_t *bfstate);
static void  storeBfUpdateOffset (bfstate_t *bfstate);
static void  storeBfOrderKick (bfstate_t *bfstate);
static char *storeBfOpDump (bfop_t *op);
static char *storeBfStateDump (bfstate_t *bfstate);
static void  storeBfIOFreeEntry(void *sio);


storeIOState *
storeBfOpen(SwapDir * sd, StoreEntry * e, STFNCB * file_callback,
	    STIOCB * callback, void *callback_data)
{
    storeIOState *sio;
    bfstate_t *bfstate;
    bfinfo_t *bfi = sd->fsdata;
    bfop_t *op;

    debug(81, 3) ("storeBfOpen: key %s\n", storeKeyText(e->key));

    sio = CBDATA_ALLOC(storeIOState, storeBfIOFreeEntry);
    sio->fsstate = bfstate = memPoolAlloc(bf_state_pool);
    sio->swap_dirn = sd->index;
    sio->mode = O_RDONLY;
    sio->callback = callback;
    sio->callback_data = callback_data;
    sio->e = e;
    cbdataLock(callback_data);

    storeBfDirFillRa(e, &bfstate->ra);
    bfstate->fd = -1;		/* not opened */
    bfstate->errflag = 0;
    bfstate->ops.next = bfstate->ops.prev = &bfstate->ops;
    op = &bfstate->ops;
    op->type = BFOPEN;
    op->sio = sio;
    op->aio.aio_fildes = bfi->dir;
    op->aio.aio_buf = &bfstate->ra;
    op->aio.aio_nbytes = REISERFS_IOC_RAWOPEN;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;
    
    if (storeBfLaunch (sd, aio_ioctl, op) != 0) {
	cbdataUnlock(callback_data);
	cbdataFree(sio);
	return NULL;
    }
    return sio;
}


storeIOState *
storeBfCreate(SwapDir * sd, StoreEntry * e, STFNCB * file_callback,
    STIOCB * callback, void *callback_data)
{
    storeIOState *sio;
    bfstate_t *bfstate;
    bfinfo_t *bfi = sd->fsdata;
    bfop_t *op;

    debug(81, 3) ("storeBfCreate: key %s flags %s\n",
		  storeKeyText(e->key), storeEntryFlags (e));

    /* Avoid disk subsystem overload.  May lead to losing hits, so
       don't set the limit too low. */
    if (bf_stats.away > bfi->refuse_swapout) {
	bf_stats.swapout_refused++;
	return NULL;
    }

    sio = CBDATA_ALLOC(storeIOState, storeBfIOFreeEntry);
    sio->fsstate = bfstate = memPoolAlloc(bf_state_pool);
    sio->swap_dirn = sd->index;
    sio->mode = O_WRONLY | O_CREAT | O_TRUNC; /* if that meant something :) */
    sio->callback = callback;
    sio->callback_data = callback_data;
    sio->e = e;
    cbdataLock(callback_data);

    storeBfDirFillRa(e, &bfstate->ra);
    bfstate->fd = -1;		/* not opened */
    bfstate->errflag = 0;
    bfstate->ops.next = bfstate->ops.prev = &bfstate->ops;
    op = &bfstate->ops;
    op->type = BFCREAT;
    op->sio = sio;
    op->aio.aio_fildes = bfi->dir;
    op->aio.aio_buf = &bfstate->ra;
    op->aio.aio_nbytes = REISERFS_IOC_RAWCREAT;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;

    /* Do not launch anything now; will lauch creat and subsequent
       writes when close request arrives */

    /* storeBfDirReplAdd(sd, e); */
    return sio;
}


void
storeBfClose(SwapDir * sd, storeIOState * sio)
{
    bfstate_t *bfstate = sio->fsstate;

    debug(81, 3) ("storeBfClose: dirno %d, key %s\n", sd->index,
		  storeKeyText (sio->e->key));

    bfstate->close_request = 1;

    /* Creat and subsequent writes are launched all at once when we
       have them all queued.  Now. */
    if (bfstate->ops.type == BFCREAT) {
	bfop_t *op = &bfstate->ops;
	if (storeBfLaunch (sd, aio_ioctl, op) != 0) {
	    storeBfIOAbort (op->sio, op->type, errno);
	    storeBfKickQueue (bfstate);
	}
	return;
    }
    
    storeBfTryLaunchClose (sio);
}


void
storeBfRead(SwapDir * sd, storeIOState * sio, char *buf, size_t size,
	    off_t offset, STRCB * callback, void *callback_data)
{
    bfstate_t *bfstate = sio->fsstate;
    bfop_t *op;

    op = memPoolAlloc(bf_op_pool);
    op->type = BFREAD;
    op->sio = sio;
    op->callback.read = callback;
    op->callback_data = callback_data;
    op->aio.aio_buf = buf;
    op->aio.aio_nbytes = size;
    op->aio.aio_offset = offset;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;

    /* Now that we have remembered it, we don't want it to be freed */
    cbdataLock(callback_data);

    debug(81, 3) ("storeBfRead: %s", storeBfOpDump(op));

    storeBfOpEnqueue (bfstate, op);
    storeBfKickQueue (bfstate);
}


void
storeBfWrite(SwapDir * sd, storeIOState * sio, char *buf, size_t size,
	     off_t offset, FREE * free_func)
{
    bfstate_t *bfstate = sio->fsstate;
    bfop_t *op;

    op = memPoolAlloc(bf_op_pool);
    op->type = BFWRITE;
    op->sio = sio;
    op->callback.write = free_func;
    op->callback_data = NULL;
    op->aio.aio_buf = buf;
    op->aio.aio_nbytes = size;
    op->aio.aio_offset = offset;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;

    debug(81, 3) ("storeBfWrite: %s", storeBfOpDump(op));

    storeBfOpEnqueue (bfstate, op);
    /*storeBfKickQueue (bfstate);  tentative: launch bfwrite on close only */

    debug(81, 8) ("storeBfWrite:%d %s", __LINE__, storeBfStateDump (bfstate));
}


void
storeBfUnlink(SwapDir *sd, StoreEntry *e)
{
    debug(81, 3) ("storeBfUnlink: dirno %d, key %s\n",
		  sd->index, storeKeyText(e->key));
    /* storeBfDirReplRemove(e); */
    storeBfUnlink2 (sd, e, NULL);
}

static int
storeBfUnlink2(SwapDir *sd, StoreEntry *e, storeIOState *sio)
{
    reiserfs_raw_ioctl_arg *ra;
    bfinfo_t *bfi = sd->fsdata;
    bfop_t *op;

    /* For Butterfly unlink is rare operation as cache replacement is
       carried out in the kernel.  So we can afford to xcalloc to
       avoid complicating our data */

    ra = xcalloc (1, sizeof (reiserfs_raw_ioctl_arg));
    storeBfDirFillRa(e, ra);

    op = memPoolAlloc(bf_op_pool);
    op->next = op->prev = op;
    op->type = BFUNLINK;
    op->sio = sio;
    op->aio.aio_fildes = bfi->dir;
    op->aio.aio_buf = ra;
    op->aio.aio_nbytes = REISERFS_IOC_RAWUNLINK;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;

    if (storeBfLaunch (sd, aio_ioctl, op) != 0) {
	if (errno == EAGAIN) {
	    int x = ioctl (bfi->dir, REISERFS_IOC_RAWUNLINK, ra);
	    debug(81, 1) ("storeBfUnlink: resorted to sync unlink, =%d\n", x);
	}
	xfree (ra);
	memPoolFree(bf_op_pool, op);
	return -1;
    }

    /*
     * Note: we leave op&ra non-freed.  Will get op back in
     * sigev_value upon completion and free then.
     */

    return 0;
}


void
storeBfSetattr(SwapDir *sd, StoreEntry * e)
{
    reiserfs_raw_ioctl_arg *ra;
    bfinfo_t *bfi = sd->fsdata;
    bfop_t *op;

    debug(81, 3) ("storeBfSetattr: dirno %d, key %s\n",
		  sd->index, storeKeyText(e->key));

    ra = xcalloc (1, sizeof (reiserfs_raw_ioctl_arg));
    storeBfDirFillRa(e, ra);

    op = memPoolAlloc(bf_op_pool);
    op->next = op->prev = op;
    op->type = BFSETATTR;
    op->sio = NULL;
    op->aio.aio_fildes = bfi->dir;
    op->aio.aio_buf = ra;
    op->aio.aio_nbytes = REISERFS_IOC_RAWSETATTR;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;

    if (storeBfLaunch (sd, aio_ioctl, op) != 0) {
	/* OK, maybe next time... */
	xfree (ra);
	memPoolFree(bf_op_pool, op);
	return;
    }

    /*
     * Note: we leave op&ra non-freed.  Will get op back in
     * sigev_value upon completion and free then.
     */
}


/* Wrapper for aio_*, provides debugging convenience and simplifies
   accounting */

static int
storeBfLaunch (SwapDir * sd, int (*aiofunc)(aiocb *), bfop_t *op)
{
    int x;
    struct bf_stats_counters *c = &bf_stats.counters[op->type];

    debug(81, 2) ("storeBfLaunch: %s", storeBfOpDump(op));

    if (bf_opstats)
	op->aio.aio_times[AIOTIME_REQUEST] = kaio_time ();

    x = aiofunc (&op->aio);
    c->ops++;
    if (x < 0) {
	c->fail++;
	if (errno == EAGAIN) bf_stats.eagain++;
	debug(50, 1) ("storeBfLaunch: type %d: %s\n", op->type, xstrerror());
    }
    else {
	op->started = 1;
	op->starttime = tvMsec (current_time);

	op->busychain = NULL;
	*bf_busyops[op->type].tail = op;
	bf_busyops[op->type].tail = &op->busychain;

	if (++bf_stats.away > bf_stats.max_away)
	    bf_stats.max_away = bf_stats.away;
    }

    return x;
}


/* Start all queued i/o operations that are ready to be launched.  It
   it OK and even good to start multiple reads or writes in parallel.
   Returns -1 if launch attempts hit EAGAIN, 0 otherwise */

int
storeBfKickQueue (bfstate_t *bfstate)
{
    storeIOState *sio = bfstate->ops.sio;
    bfop_t *op;
    int x;

    /*
     * Remove completed and aborted ops off the head of the queue.
     */
    storeBfSweep (bfstate);

    /*
     * If close requested and q is empty, launch the close 
     */
    if (bfstate->close_request && bfstate->ops.next == &bfstate->ops) {
	storeBfTryLaunchClose (sio);
	return 0;
    }

    /*
     * If no fd, we wannot start read/write yet, we're waiting for
     * OPEN/CREAT to finish.  storeBfOpCompletion will call
     * storeBfKickQueue again, and then...
     */
    if (bfstate->fd < 0)
	return 0;

    /*
     * If we are shutting down after error, don't start new
     * operations.
     */
    if (bfstate->aborting)
	/*
	 * Now we either have flying ops in the q, or close launched.
	 * So OpCompletion is guaranteed, just return and wait for it.
	 */
	return 0;

    /*
     * At last, our main duty.  Launch everything launchable.
     */
    for (op = bfstate->ops.next; op != &bfstate->ops; op = op->next) {
	if (op->started)
	    continue;

	op->aio.aio_fildes = bfstate->fd;

	switch (op->type) {
	case BFREAD:
	    x = storeBfLaunch (NULL, aio_read, op);
	    break;
	case BFWRITE:
	    x = storeBfLaunch (NULL, aio_write, op);
	    break;
	default:
	    abort ();
	}

	if (x < 0) {
	    if (errno == EAGAIN) {
		/* Looks like too many ops are launched already.  Wait
                   for some to land before continuing.  The nasty
                   thing is that if no ops are currently in flight,
                   nobody will call us again (except the user -- but
                   that's unreliable).  Therefore, ask for help from
                   upper layers (which have its price, follow me...) */
		storeBfOrderKick (bfstate);
		return -1;
	    }
	    else {
		/* Real error.  This is the end of the sio.  Raise the
                   abort flag and recurse -- as the op failed here,
                   not in OpCompletion, we must do funeral
                   arrangements ourselves  */
		storeBfIOAbort (op->sio, op->type, errno);
		storeBfKickQueue (bfstate);
		return 0;
	    }
	}

#if 0
	/* tentative: "don't overdo the asyncronous operations", as
	   Henrik says.  Write is already asynchronous by nature, so
	   if it does not return fast then launching more writes will
	   only clutter the queue without any speed gain */
	if (op->type == BFWRITE)
	    break;
#endif
    }

    return 0;
}


/* Handle async I/O operation completion event. 'op' is the descriptor
   of the completed operation. */

void    
storeBfOpCompletion (bfop_t *op)
{
    storeIOState *sio = op->sio;
    bfstate_t *bfstate = sio? sio->fsstate: NULL;
    struct bf_stats_counters *c;
    int x = aio_error (&op->aio);
    int dura = tvMsec (current_time) - op->starttime;
    int rlen;

    op->finished = 1;

    debug(81, 2) ("storeBfOpCompletion: %d ms, x=%d %s",
		  dura, x, storeBfOpDump(op));

    /* assert (! sio || cbdataValid(sio)), because we never call
       storeBfIOCallback before the file is really closed. */

    if (op->type == BFNONE || op->type >= BFMAX) {
	debug(81, 1) ("storeBfOpCompletion: bad op type (%d)\n", op->type);
	/* this "cannot happen" so error handling is sloppy */
	return;
    }

    /*
     * Update statistics counters
     */
    c = &bf_stats.counters[op->type];
    (x == 0) ? c->success++ : c->fail++;
    if (c->max_dura < dura)
	c->max_dura = dura;
    c->avg_dura = (c->avg_dura * (c->ops-1) + dura) / c->ops;
    c->avg_dura100 = (c->avg_dura100 * 99 + dura) / 100;

    /*
     * Do operation-specific processing
     */
    switch (op->type) {
    case BFCREAT:
	if (x == EEXIST) {
	    /* id collision (not too likely) or REFRESH_MISS.  Handle
	       by removing the old file and retrying */
	    debug(81, 2) ("storeBfCompletion: BFCREAT: dir%d, probably id collision!\n",
			  sio->swap_dirn, x);
	    if (bfstate->unlink_attempted)
		debug(81, 1) ("storeBfCompletion: BFCREAT: dir%d, lingering unlink %s\n",
			      sio->swap_dirn, storeKeyText(sio->e->key));
	    else {
		int x = storeBfUnlink2 (INDEXSD (sio->swap_dirn), sio->e, sio);
		if (x >= 0) {
		    bfstate->unlink_attempted++;
		    return;
		}
	    }
	    /* unlink failed, continue processing the error (this will
               result in SWAPOUT_FAIL) */
	}
    create_fail:
	/* fall-through */

    case BFOPEN:
	if (x != 0)
	    storeBfIOAbort (sio, op->type, x);
	else {
	    bfstate->fd = aio_return (&op->aio);
	    sio->e->swap_file_sz = bfstate->ra.size;
	    store_open_disk_fd++;
	    debug(81, 4) ("storeBfOpCompletion: sio %p fd %d opened\n", sio, bfstate->fd);
	}
	break;

    case BFREAD:
	if (x != 0)
	    storeBfIOAbort (sio, op->type, x);
	rlen = (x == 0)? aio_return (&op->aio): -1;
	op->rlen = (x == 0)? aio_return (&op->aio): 0;
	storeBfUpdateOffset (bfstate);
	if (! bfstate->aborting && cbdataValid(op->callback_data))
	    op->callback.read (op->callback_data, op->aio.aio_buf, rlen);
	cbdataUnlock (op->callback_data);
	break;

    case BFWRITE:
	if (x != 0)
	    storeBfIOAbort (sio, op->type, x);
	op->rlen = (x == 0)? aio_return (&op->aio): 0;
	storeBfUpdateOffset (bfstate);
	if (op->callback.write)
	    op->callback.write (op->aio.aio_buf); /* free_func */
	break;

    case BFCLOSE:
	debug(81, 4) ("storeBfOpCompletion: fd %d closed (x=%d)\n", bfstate->fd, x);
	if (x != 0)
	    storeBfCloseLastResort (sio);
	else {
	    bfstate->fd = -1;
	    store_open_disk_fd--;
	}
	storeBfIOCallback (sio, bfstate->errflag);
	/* That's the showdown.  sio is freed by now */
	return;

    case BFUNLINK:
	xfree (op->aio.aio_buf); /* ra */
	memPoolFree(bf_op_pool, op);
	if (sio) {
	    /* This unlink was launched to handle EEXIST error.
	       Now retry CREATE */
	    op = &bfstate->ops;
	    assert (op->type == BFCREAT);
	    if (x) {
		x = EEXIST;
		goto create_fail;
	    }
	    op->started = op->finished = 0;
	    if (storeBfLaunch (NULL, aio_ioctl, op) != 0) {
		x = EEXIST;
		goto create_fail;
	    }
	}
	return;

    case BFSETATTR:
	xfree (op->aio.aio_buf); /* ra */
	memPoolFree(bf_op_pool, op);
	return;

    default:
	abort ();		/* cannot happen. */
    }

    /*
     * Ok, op done.  Go ask for more...
     */
    storeBfKickQueue (bfstate);

    /*
     * Don't use op below this point.  storeBfKickQueue calls
     * storeBfSweep that potentially destroys the bfop pointed by it.
     */

    return;
}



/*  === STATIC =========================================================== */



/* Launch async file close operation.  We are one step away from
   storeBfIOCallback now.  The queue must be empty before this is
   called.  Note that the call to this possibly results in sio
   invalidation, so after calling this, return immediately; don't use
   any pointers like op or bfstate that you might have. */

static void
storeBfTryLaunchClose (storeIOState *sio)
{
    bfstate_t *bfstate = sio->fsstate;
    bfop_t *op = &bfstate->ops;

    if (bfstate->close_started)
	return;
    if (bfstate->ops.next != &bfstate->ops)
	return;			/* we'll be called again when q empties */

    debug(81, 3) ("storeBfTryLaunchClose: %s", storeBfStateDump (bfstate));

    if (bfstate->fd < 0) {
	/* not opened -- no need to close.  Bye right now */
	storeBfIOCallback (sio, bfstate->errflag);
	return;
    }
    op->type = BFCLOSE;
    op->sio = sio;
    op->aio.aio_fildes = bfstate->fd;
    op->aio.aio_sigevent.sigev_notify = SIGEV_NONE;
    op->aio.aio_sigevent.sigev_signo = 0;
    op->aio.aio_sigevent.sigev_value.sival_ptr = op;
    
    debug(81, 2) ("storeBfTryLaunchClose: fd=%d %s", bfstate->fd, storeBfOpDump(op));
    debug(81, 2) ("storeBfTryLaunchClose: callers: %p %p %p %p\n",
		  __builtin_return_address(0),
		  __builtin_return_address(1),
		  __builtin_return_address(2),
		  __builtin_return_address(3));

    bfstate->close_started = 1;
    if (storeBfLaunch (NULL, aio_close, op) < 0) {
	storeBfCloseLastResort (sio);
	storeBfIOCallback (sio, bfstate->errflag);
    }
}


/* Called if aio_close failed, to attempt synchronous file close.  This
   will never be called ;-) 'errflag' is the error returned by
   aio_close. */

static void
storeBfCloseLastResort (storeIOState *sio)
{
    bfstate_t *bfstate = sio->fsstate;

    debug(81, 1) ("Resorting to sync close()...\n");

    if (close (bfstate->fd) != 0) 
	debug(50, 1) ("close: %s\n", xstrerror());
    else {
	bfstate->fd = -1;
	store_open_disk_fd--;
    }
}


/* Call the STIOCB aka "file closed" callback.  This is the final
   accord in the life of every sio */

static void
storeBfIOCallback (storeIOState *sio, int errflag)
{
    bfstate_t *bfstate = sio->fsstate;
    int valid = cbdataValid(sio->callback_data);

    debug(81, 3) ("storeBfIOCallback: errflag=%d\n", errflag);
    assert (bfstate->ops.next == &bfstate->ops);
    cbdataUnlock(sio->callback_data);
    if (valid) {
	errno = errflag;
	sio->callback(sio->callback_data, errflag? DISK_ERROR: DISK_OK, sio);
    }
    cbdataFree(sio);
}


/* Change the sio state to "aborting" -- that is, arrange to shut down
   all I/O active on the sio, close the file and eventually call
   storeBfIOCallback.  This is called if any operation (except close)
   fails.  It does no actions, just sets appropriate flags.
   storeBfKickQueue will notice and shutdown instead of continuing. */

static void
storeBfIOAbort (storeIOState *sio, int type, int errflag)
{
    bfstate_t *bfstate = sio->fsstate;

    debug(50, 3) ("storeBfIOAbort: sio %p, aio_{%d}, errflag=%d\n",
		  sio, type, errflag);

    if (errflag && !bfstate->errflag)
	bfstate->errflag = errflag;
    bfstate->close_request = 1;
    bfstate->aborting = 1;

    /* Don't sweep here -- this will invalidate ops while our callers
       are often interesed in them, despite the aborting.
       storeBfKickQueue will do that. */

    if (bfstate->ops.next != &bfstate->ops) {
	/* TODO: aio_cancel all ops currently in flight? */
    }
}


/* Add op to the tail of the op queue */

static void
storeBfOpEnqueue (bfstate_t *bfstate, bfop_t *op)
{
    if (bfstate->close_started) {
	/* Too late, we are closing. */
	if (op->callback_data)
	    cbdataUnlock (op->callback_data);
	memPoolFree (bf_op_pool, op);
    }
    op->prev = bfstate->ops.prev;
    bfstate->ops.prev = op;
    op->next = &bfstate->ops;
    op->prev->next = op;
}


/* Remove finished ops from the start of the list.  Finished ops
   interspersed with unfinished ones are left alone, they will be
   needed for future storeBfUpdateOffset()s.  If aborting, kill
   unstarted ops too. */

static void
storeBfSweep (bfstate_t *bfstate)
{
    bfop_t *op;
    bfop_t *next;

    for (op = bfstate->ops.next; op != &bfstate->ops; op = next) {
	next = op->next;
	if (op->finished || (bfstate->aborting && ! op->started)) {
	    op->prev->next = op->next;
	    op->next->prev = op->prev;
	    if (! op->finished && op->callback_data)
		cbdataUnlock (op->callback_data);
	    memPoolFree (bf_op_pool, op);
	    continue;
	}
	break;
    }
}


/* Calculate file offset for storeOffset by examining 'finished' ops
   in the queue.  We rely on the fact that queue is ordered by offset,
   and there is for sure at least one element in the queue (the one
   that storeBfOpCompletion processes when it calls us). */

static void
storeBfUpdateOffset (bfstate_t *bfstate)
{
    storeIOState *sio = bfstate->ops.sio;
    bfop_t *op;

    for (op = bfstate->ops.next; op != &bfstate->ops; op = op->next) {
	if (op->finished) {
	    off_t offset = op->aio.aio_offset + op->rlen;
	    if (sio->offset < offset)
		sio->offset = offset;
	    continue;
	}
	break;
    }

    debug(81, 5) ("storeBfUpdateOffset: sio %p, offset %lu\n",
		  sio, sio->offset);
}


/* Arrange for our queue being kicked externally at a later time.
   Called when we lose initiative because of EAGAIN. */

static void
storeBfOrderKick (bfstate_t *bfstate)
{
    storeIOState *sio = bfstate->ops.sio;
    bfinfo_t *bfi = INDEXSD(sio->swap_dirn)->fsdata;
    bfop_t *op;

    /* Avoid the external help if possible.  Check if there are ops in
       flight; if yes, their completion will do the kick and no
       external help is needed. */

    for (op = bfstate->ops.next; op != &bfstate->ops; op = op->next) {
	if (op->started && ! op->finished)
	    /* found! */
	    return;
    }

    /* No way.  Go begging.  We have to use usual callback calling
       conventions, because user can kick us before the ordered kick
       is activated, and when it at last activated it might find sio
       freed.  */

    cbdataLock (sio);
    linklistPush(&(bfi->kicklater), sio);
}

/* Debugging aid */

static char *
storeBfOpDump (bfop_t *op)
{
    static char buf[1000];
    static char *optypestr[] = {
	"BFNONE", "BFOPEN", "BFCREAT", "BFREAD", "BFWRITE",
	"BFCLOSE", "BFUNLINK", "BFSETATTR",
    };
    char *typestr = op->type >= BFMAX? "bad!": optypestr[op->type];
    int n = 0;

    n += snprintf (buf+n, sizeof (buf)-n,
		   "op %p: %9s sio %p %s %s",
		   op, typestr, op->sio,
		   op->started? "started": "",
		   op->finished? "finished": "");
    if (op->type == BFREAD || op->type == BFWRITE)
	n += snprintf (buf+n, sizeof (buf)-n,
		       " %u@%lu",
		       op->aio.aio_nbytes,
		       op->aio.aio_offset);
    if (op->type == BFWRITE)
	n += snprintf (buf+n, sizeof (buf)-n,
		       " %02x %02x %02x %02x %02x",
		       ((unsigned char *)op->aio.aio_buf)[0],
		       ((unsigned char *)op->aio.aio_buf)[1],
		       ((unsigned char *)op->aio.aio_buf)[2],
		       ((unsigned char *)op->aio.aio_buf)[3],
		       ((unsigned char *)op->aio.aio_buf)[4]);
    n += snprintf (buf+n, sizeof (buf)-n, "\n");
    return buf;
}


static char *
storeBfStateDump (bfstate_t *bfstate)
{
    bfop_t *op;
    static char buf[10000];
    char *p = buf;

    p += snprintf (p, &buf[10000] - p, "bfstate %p\n", bfstate);
    p += snprintf (p, &buf[10000] - p, " fd %d errflag %d %s %s %s\n",
		  bfstate->fd,
		  bfstate->errflag,
		  (bfstate->close_request? "CR":""),
		  (bfstate->close_started? "CS":""),
		  (bfstate->aborting? "ABORTING":""));
    op = &bfstate->ops;
    do {
	p += snprintf (p, &buf[10000] - p, storeBfOpDump (op));
	op = op->next;
    } while (op != &bfstate->ops);

    return buf;
}

static void
storeBfIOFreeEntry(void *_sio)
{
    storeIOState * sio = _sio;
    bfstate_t *bfstate = sio->fsstate;

    /* queue must be freed before cbdataFree()ing the sio */
    assert (bfstate->ops.next == &bfstate->ops);

    memPoolFree(bf_state_pool, sio->fsstate);
}
